"""
MIT License

Copyright (c) 2022 AlgoWorld

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from pyteal import Mode, compileTeal

from algoworld_contracts.auction.clear import clear
from algoworld_contracts.auction.escrow import escrow
from algoworld_contracts.auction.manager import manager
from algoworld_contracts.auction.proxy import proxy
from algoworld_contracts.swapper.asa_to_asa_swapper import AsaToAsaSwapConfig, swapper
from algoworld_contracts.swapper.asas_to_algo_swapper import (
    AsasToAlgoSwapConfig,
    multi_asa_swapper,
)
from algoworld_contracts.swapper.swap_proxy import SwapProxy, swapper_proxy

TEAL_VERSION = 6


def get_clear_teal():
    return compileTeal(clear(), Mode.Application, version=TEAL_VERSION)


def get_escrow_teal(app_id: int, nft_id: int, fee_address_a: str, fee_address_b: str):
    return compileTeal(
        escrow(app_id, nft_id, fee_address_a, fee_address_b),
        Mode.Signature,
        version=TEAL_VERSION,
    )


def get_proxy_teal(proxy_id: int):
    return compileTeal(proxy(proxy_id), Mode.Signature, version=TEAL_VERSION)


def get_manager_teal(fee_address_a: str, fee_address_b: str, contract_version: str):
    return compileTeal(
        manager(fee_address_a, fee_address_b, contract_version),
        Mode.Application,
        version=TEAL_VERSION,
    )


def get_swapper_teal(
    swap_creator: str,
    offered_asa_id: int,
    offered_asa_amount: int,
    requested_asa_id: int,
    requested_asa_amount: int,
    incentive_fee_address: str,
    incentive_fee_amount: int,
):
    return compileTeal(
        swapper(
            AsaToAsaSwapConfig(
                swap_creator=swap_creator,
                offered_asa_id=offered_asa_id,
                offered_asa_amount=offered_asa_amount,
                requested_asa_id=requested_asa_id,
                requested_asa_amount=requested_asa_amount,
                incentive_fee_address=incentive_fee_address,
                incentive_fee_amount=incentive_fee_amount,
            )
        ),
        Mode.Signature,
        version=TEAL_VERSION,
    )


def get_swapper_proxy_teal(swap_creator: str, version: str):
    return compileTeal(
        swapper_proxy(SwapProxy(swap_creator, version)),
        Mode.Signature,
        version=TEAL_VERSION,
    )


def get_multi_swapper_teal(
    swap_creator: str,
    offered_asa_amounts: dict[int, int],
    requested_algo_amount: int,
    max_fee: int,
    optin_funding_amount: int,
    incentive_fee_address: str,
    incentive_fee_amount: int,
):
    return compileTeal(
        multi_asa_swapper(
            AsasToAlgoSwapConfig(
                swap_creator=swap_creator,
                offered_asa_amounts=offered_asa_amounts,
                requested_algo_amount=requested_algo_amount,
                max_fee=max_fee,
                optin_funding_amount=optin_funding_amount,
                incentive_fee_address=incentive_fee_address,
                incentive_fee_amount=incentive_fee_amount,
            )
        ),
        Mode.Signature,
        version=TEAL_VERSION,
    )



"""
MIT License

Copyright (c) 2022 AlgoWorld

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import dataclasses
import sys

from pyteal import Addr, And, Cond, Expr, Global, Gtxn, Int, Mode, TxnType, compileTeal

from algoworld_contracts.common.utils import parse_params

"""
Multi ASA to ALGO Atomic Swapper
1. Multi ASA Opt-In
2. Offered Multi ASA / Requested ALGO Swap
3. Close Multi ASA Swap
"""

TEAL_VERSION = 6
BASE_OPTIN_FUNDING_AMOUNT = 210000


@dataclasses.dataclass
class AsasToAlgoSwapConfig:
    swap_creator: str
    offered_asa_amounts: dict[str, int]
    requested_algo_amount: int
    max_fee: int
    optin_funding_amount: int
    incentive_fee_address: str
    incentive_fee_amount: int

    def __post_init__(self):
        assert len(self.offered_asa_amounts) <= 5
        self.body_size = len(self.offered_asa_amounts)

        # MULTI ASA OPTIN
        self.optin_header = {"fee": 0}
        self.optin_bottom = {}
        self.optin_gsize = (
            len(self.optin_header) + self.body_size + len(self.optin_bottom)
        )

        # MULTI ASA SWAP
        self.swap_header = {"incentive_fee": 0, "requested_algo_xfer": 1}
        self.swap_bottom = {}
        self.swap_gsize = len(self.swap_header) + self.body_size + len(self.swap_bottom)

        # CLOSE MULTI ASA SWAP
        self.close_swap_header = {}
        self.close_swap_bottom = {
            "close_out": self.body_size + 0,
            "proof": self.body_size + 1,
        }
        self.close_swap_gsize = (
            len(self.close_swap_header) + self.body_size + len(self.close_swap_bottom)
        )


def multi_asa_swapper(cfg: AsasToAlgoSwapConfig) -> Expr:

    multi_asa_optin_type_check = [
        Gtxn[len(cfg.optin_header) + asa].type_enum() == TxnType.AssetTransfer
        for asa in range(cfg.body_size)
    ]

    is_multi_asa_optin = And(
        Global.group_size() == Int(cfg.optin_gsize),
        Gtxn[cfg.optin_header["fee"]].type_enum() == TxnType.Payment,
        *multi_asa_optin_type_check,
    )

    multi_asa_xfer_type_check = [
        Gtxn[len(cfg.swap_header) + asa].type_enum() == TxnType.AssetTransfer
        for asa in range(cfg.body_size)
    ]

    is_multi_asa_swap = And(
        Global.group_size() == Int(cfg.swap_gsize),
        Gtxn[cfg.swap_header["incentive_fee"]].type_enum() == TxnType.Payment,
        Gtxn[cfg.swap_header["requested_algo_xfer"]].type_enum() == TxnType.Payment,
        *multi_asa_xfer_type_check,
    )

    multi_asa_close_type_check = [
        Gtxn[len(cfg.close_swap_header) + asa].type_enum() == TxnType.AssetTransfer
        for asa in range(cfg.body_size)
    ]

    is_close_swap = And(
        Global.group_size() == Int(cfg.close_swap_gsize),
        *multi_asa_close_type_check,
        Gtxn[cfg.close_swap_bottom["close_out"]].type_enum() == TxnType.Payment,
        Gtxn[cfg.close_swap_bottom["proof"]].type_enum() == TxnType.Payment,
    )

    return Cond(
        [is_multi_asa_optin, multi_asa_optin(cfg)],
        [is_multi_asa_swap, multi_asa_swap(cfg)],
        [is_close_swap, multi_asa_close_swap(cfg)],
    )


def multi_asa_optin(cfg: AsasToAlgoSwapConfig):

    # Code repetition of `for asa in range(cfg.body_size)` for every check
    # has been kept in favor of readability. A unique for loop would be more
    # succint but maybe harder to read.
    # The following check implies all the ASA xfers have the same sender
    multi_asa_optin_senders = [
        Gtxn[len(cfg.optin_header) + asa].sender()
        == Gtxn[cfg.optin_header["fee"]].receiver()
        for asa in range(cfg.body_size)
    ]

    asa = 0
    multi_asa_optin_xfer_asset = []
    for k in cfg.offered_asa_amounts:
        multi_asa_optin_xfer_asset += [
            Gtxn[len(cfg.optin_header) + asa].xfer_asset() == Int(int(k))
        ]
        asa += 1

    multi_asa_optin_assets_receivers = [
        Gtxn[len(cfg.optin_header) + asa].sender()
        == Gtxn[len(cfg.optin_header) + asa].asset_receiver()
        for asa in range(cfg.body_size)
    ]

    multi_asa_optin_assets_amounts = [
        Gtxn[len(cfg.optin_header) + asa].asset_amount() == Int(0)
        for asa in range(cfg.body_size)
    ]

    return And(
        Gtxn[cfg.optin_header["fee"]].sender() == Addr(cfg.swap_creator),
        Gtxn[cfg.optin_header["fee"]].amount() >= Int(cfg.optin_funding_amount),
        Gtxn[cfg.optin_header["fee"]].fee() <= Int(cfg.max_fee),
        Gtxn[cfg.optin_header["fee"]].rekey_to() == Global.zero_address(),
        Gtxn[cfg.optin_header["fee"]].close_remainder_to() == Global.zero_address(),
        *multi_asa_optin_senders,
        *multi_asa_optin_xfer_asset,
        *multi_asa_optin_assets_receivers,
        *multi_asa_optin_assets_amounts,
    )


def multi_asa_swap(cfg: AsasToAlgoSwapConfig):

    offered_multi_asa_xfer_max_fee = [
        Gtxn[len(cfg.swap_header) + asa].fee() <= Int(cfg.max_fee)
        for asa in range(cfg.body_size)
    ]

    offered_multi_asa_xfer_rekey_to = [
        Gtxn[len(cfg.swap_header) + asa].rekey_to() == Global.zero_address()
        for asa in range(cfg.body_size)
    ]

    offered_multi_asa_xfer_asset_sender = [
        Gtxn[len(cfg.swap_header) + asa].asset_sender() == Global.zero_address()
        for asa in range(cfg.body_size)
    ]

    offered_multi_asa_xfer_asset_close_to = [
        Gtxn[len(cfg.swap_header) + asa].asset_close_to() == Global.zero_address()
        for asa in range(cfg.body_size)
    ]

    asa = 0
    offered_multi_asa_xfer_asset_ids = []
    offered_multi_asa_xfer_asset_amounts = []
    for k, v in cfg.offered_asa_amounts.items():
        offered_multi_asa_xfer_asset_ids += [
            Gtxn[len(cfg.swap_header) + asa].xfer_asset() == Int(int(k))
        ]
        offered_multi_asa_xfer_asset_amounts += [
            Gtxn[len(cfg.swap_header) + asa].asset_amount() == Int(int(v))
        ]
        asa += 1

    offered_multi_asa_xfer_asset_receiver = [
        Gtxn[len(cfg.swap_header) + asa].asset_receiver()
        == Gtxn[cfg.swap_header["requested_algo_xfer"]].sender()
        for asa in range(cfg.body_size)
    ]

    offered_multi_asa_xfer_precondition = And(
        *offered_multi_asa_xfer_max_fee,
        *offered_multi_asa_xfer_rekey_to,
        *offered_multi_asa_xfer_asset_sender,
        *offered_multi_asa_xfer_asset_close_to,
    )

    return And(
        offered_multi_asa_xfer_precondition,
        Gtxn[cfg.swap_header["incentive_fee"]].receiver()
        == Addr(cfg.incentive_fee_address),
        Gtxn[cfg.swap_header["incentive_fee"]].amount()
        == Int(cfg.incentive_fee_amount),
        Gtxn[cfg.swap_header["requested_algo_xfer"]].amount()
        == Int(cfg.requested_algo_amount),
        Gtxn[cfg.swap_header["requested_algo_xfer"]].receiver()
        == Addr(cfg.swap_creator),
        *offered_multi_asa_xfer_asset_ids,
        *offered_multi_asa_xfer_asset_amounts,
        *offered_multi_asa_xfer_asset_receiver,
    )


def multi_asa_close_swap(cfg: AsasToAlgoSwapConfig):

    multi_asa_close_max_fee = [
        Gtxn[len(cfg.close_swap_header) + asa].fee() <= Int(cfg.max_fee)
        for asa in range(cfg.body_size)
    ]

    multi_asa_close_rekey_to = [
        Gtxn[len(cfg.close_swap_header) + asa].rekey_to() == Global.zero_address()
        for asa in range(cfg.body_size)
    ]

    multi_asa_close_asset_sender = [
        Gtxn[len(cfg.close_swap_header) + asa].asset_sender() == Global.zero_address()
        for asa in range(cfg.body_size)
    ]

    asa = 0
    multi_asa_close_asset_ids = []
    for k, v in cfg.offered_asa_amounts.items():
        multi_asa_close_asset_ids += [
            Gtxn[len(cfg.close_swap_header) + asa].xfer_asset() == Int(int(k))
        ]
        asa += 1

    multi_asa_close_asset_receiver = [
        Gtxn[len(cfg.close_swap_header) + asa].asset_receiver()
        == Addr(cfg.swap_creator)
        for asa in range(cfg.body_size)
    ]

    multi_asa_close_asset_close_to = [
        Gtxn[len(cfg.close_swap_header) + asa].asset_close_to()
        == Addr(cfg.swap_creator)
        for asa in range(cfg.body_size)
    ]

    asa_close_precondition = And(
        *multi_asa_close_max_fee,
        *multi_asa_close_rekey_to,
        *multi_asa_close_asset_sender,
    )

    swap_close_precondition = And(
        Gtxn[cfg.close_swap_bottom["close_out"]].fee() <= Int(cfg.max_fee),
        Gtxn[cfg.close_swap_bottom["close_out"]].rekey_to() == Global.zero_address(),
    )

    return And(
        asa_close_precondition,
        swap_close_precondition,
        *multi_asa_close_asset_ids,
        *multi_asa_close_asset_receiver,
        *multi_asa_close_asset_close_to,
        Gtxn[cfg.close_swap_bottom["close_out"]].receiver() == Addr(cfg.swap_creator),
        Gtxn[cfg.close_swap_bottom["close_out"]].close_remainder_to()
        == Addr(cfg.swap_creator),
        Gtxn[cfg.close_swap_bottom["proof"]].sender() == Addr(cfg.swap_creator),
        Gtxn[cfg.close_swap_bottom["proof"]].receiver() == Addr(cfg.swap_creator),
        Gtxn[cfg.close_swap_bottom["proof"]].amount() == Int(0),
    )


def compile_stateless(program):
    return compileTeal(program, Mode.Signature, version=TEAL_VERSION)


if __name__ == "__main__":
    offered_asas = {"1": 10, "2": 10}

    params = {
        "swap_creator": "2ILRL5YU3FZ4JDQZQVXEZUYKEWF7IEIGRRCPCMI36VKSGDMAS6FHSBXZDQ",
        "offered_asa_amounts": offered_asas,
        "requested_algo_amount": 1_000_000,
        "max_fee": 1_000,
        "optin_funding_amount": BASE_OPTIN_FUNDING_AMOUNT * len(offered_asas),
        "incentive_fee_address": "RJVRGSPGSPOG7W3V7IMZZ2BAYCABW3YC5MWGKEOPAEEI5ZK5J2GSF6Y26A",
        "incentive_fee_amount": 10_000,
    }

    # Overwrite params if sys.argv[1] is passed
    if len(sys.argv) > 1:
        params = parse_params(sys.argv[1], params)

    print(compile_stateless(multi_asa_swap(AsasToAlgoSwapConfig(**params))))



"""
MIT License

Copyright (c) 2022 AlgoWorld

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys

from pyteal import (
    Addr,
    And,
    App,
    Assert,
    Btoi,
    Bytes,
    Cond,
    Global,
    Gtxn,
    If,
    Int,
    Mode,
    OnComplete,
    Or,
    Return,
    ScratchSlot,
    Seq,
    TealType,
    Txn,
    TxnType,
    compileTeal,
)

from algoworld_contracts.common.utils import parse_params

if __name__ == "__main__":
    from helpers.state import GlobalState, LocalState
else:
    from .helpers.state import GlobalState, LocalState

TEAL_VERSION = 6


class ManagerContract:
    def __init__(self, fee_addr1: str, fee_addr2: str, contract_version_value: str):
        self.nft_id = GlobalState("N")
        self.escrow_address = GlobalState("E")
        self.ask_price = GlobalState("A")
        self.bids_amount = GlobalState("B")
        self.owner_address = GlobalState("O")
        self.creator_address = GlobalState("C")
        self.bid_price = LocalState("B")
        self.contract_version = GlobalState("V")
        self.fee_addr1 = fee_addr1
        self.fee_addr2 = fee_addr2
        self.contract_version_value = int(contract_version_value)

    def on_register(self):
        return Seq(
            [
                # Set default values for user
                If(
                    Global.group_size() == Int(1),
                    Seq(
                        [
                            self.bid_price.put(Int(0)),
                            Return(Int(1)),
                        ]
                    ),
                    Return(Int(0)),
                )
            ]
        )

    def on_bid(self):
        is_valid_tx = ScratchSlot()
        prev_bid = ScratchSlot()
        return Seq(
            [
                prev_bid.store(self.bid_price.get()),
                # Increase bid_price
                If(
                    Global.group_size() == Int(2),
                    Seq(
                        [
                            Assert(
                                And(
                                    Gtxn[1].receiver() == self.escrow_address.get(),
                                    Gtxn[1].sender() == Gtxn[0].sender(),
                                    Gtxn[1].type_enum() == TxnType.Payment,
                                )
                            ),
                            self.bid_price.put(self.bid_price.get() + Gtxn[1].amount()),
                            is_valid_tx.store(Int(1)),
                        ]
                    ),
                    If(
                        Global.group_size() == Int(3),
                        Seq(
                            [
                                Assert(
                                    And(
                                        Gtxn[2].sender() == self.escrow_address.get(),
                                        # Cover fee
                                        Gtxn[1].receiver() == self.escrow_address.get(),
                                        Gtxn[2].amount() != Int(0),
                                    )
                                ),
                                self.bid_price.put(
                                    self.bid_price.get() - Gtxn[2].amount(),
                                ),
                                is_valid_tx.store(Int(1)),
                            ]
                        ),
                        is_valid_tx.store(Int(0)),
                    ),
                ),
                # Decrease bid_price
                Assert(
                    is_valid_tx.load(TealType.uint64) == Int(1),
                ),
                If(
                    And(
                        self.bid_price.get() != Int(0),
                        prev_bid.load(TealType.uint64) == Int(0),
                    ),
                    Seq([self.bids_amount.put(self.bids_amount.get() + Int(1))]),
                ),
                If(
                    And(
                        self.bid_price.get() == Int(0),
                        prev_bid.load(TealType.uint64) != Int(0),
                    ),
                    Seq([self.bids_amount.put(self.bids_amount.get() - Int(1))]),
                ),
                Return(Int(1)),
            ]
        )

    def on_set_price(self):
        return Seq(
            [
                # Deposit NFT on first call of set_price
                If(
                    Global.group_size() == Int(2),
                    Seq(
                        [
                            Assert(
                                And(
                                    self.owner_address.get() == Global.zero_address(),
                                    self.ask_price.get() == Int(0),
                                    Gtxn[1].asset_receiver()
                                    == self.escrow_address.get(),
                                    Gtxn[1].type_enum() == TxnType.AssetTransfer,
                                    Gtxn[1].xfer_asset() == self.nft_id.get(),
                                    Gtxn[1].asset_amount() == Int(1),
                                    Gtxn[1].sender() == Gtxn[0].sender(),
                                )
                            ),
                            self.owner_address.put(Txn.sender()),
                        ]
                    ),
                ),
                Assert(
                    And(
                        Btoi(Txn.application_args[1]) != self.ask_price.get(),
                        self.owner_address.get() == Txn.sender(),
                    )
                ),
                self.ask_price.put(Btoi(Txn.application_args[1])),
                If(
                    Global.group_size() == Int(3),
                    Seq(
                        [
                            Assert(Gtxn[2].sender() == self.escrow_address.get()),
                            self.owner_address.put(Global.zero_address()),
                            Return(Int(1)),
                        ]
                    ),
                ),
                Assert(self.ask_price.get() != Int(0)),
                Return(Int(1)),
            ]
        )

    def on_buy_now(self):
        return Seq(
            [
                # Ensure that ask_price and owner_address are not cleared
                Assert(
                    And(
                        self.ask_price.get() != Int(0),
                        self.owner_address.get() != Global.zero_address(),
                    )
                ),
                # Reset bid offer
                If(
                    self.bid_price.get() != Int(0),
                    Seq([self.bids_amount.put(self.bids_amount.get() - Int(1))]),
                ),
                # Increase bid_price
                If(
                    Gtxn[2].receiver() == self.escrow_address.get(),
                    Seq(
                        [
                            self.bid_price.put(self.bid_price.get() + Gtxn[2].amount()),
                        ]
                    ),
                ),
                # Decrease bid_price
                If(
                    Gtxn[2].sender() == self.escrow_address.get(),
                    Seq(
                        [
                            self.bid_price.put(self.bid_price.get() - Gtxn[2].amount()),
                        ]
                    ),
                ),
                Assert(
                    And(
                        self.bid_price.get() == self.ask_price.get(),
                        # Cover fee
                        Gtxn[1].receiver() == self.escrow_address.get(),
                        # Transfer NFT to a new owner
                        Gtxn[3].sender() == self.escrow_address.get(),
                        # Transfer ALGO to the former owner
                        Gtxn[4].receiver() == self.owner_address.get(),
                        Gtxn[4].sender() == self.escrow_address.get(),
                        Gtxn[4].amount() == self.ask_price.get() * Int(95) / Int(100),
                        Gtxn[5].receiver() == Addr(self.fee_addr1),
                        Gtxn[5].sender() == self.escrow_address.get(),
                        Gtxn[5].amount()
                        == (self.ask_price.get() * Int(5) / Int(100)) / Int(2),
                        Gtxn[6].receiver() == Addr(self.fee_addr2),
                        Gtxn[6].sender() == self.escrow_address.get(),
                        Gtxn[6].amount()
                        == (self.ask_price.get() * Int(5) / Int(100)) / Int(2),
                    )
                ),
                # Reset ask offer
                self.ask_price.put(Int(0)),
                self.owner_address.put(Global.zero_address()),
                self.bid_price.put(Int(0)),
                Return(Int(1)),
            ]
        )

    def on_sell_now(self):
        return Seq(
            [
                Assert(
                    And(
                        # Make sure there's a bid in buyer's account
                        App.localGet(Int(1), Bytes("B")) != Int(0),
                        # Transfer ALGO to the seller
                        Gtxn[2].amount()
                        == App.localGet(Int(1), Bytes("B")) * Int(95) / Int(100),
                        Gtxn[2].sender() == self.escrow_address.get(),
                        # Cover fee
                        Gtxn[1].receiver() == self.escrow_address.get(),
                        Gtxn[4].receiver() == Addr(self.fee_addr1),
                        Gtxn[4].amount()
                        == (App.localGet(Int(1), Bytes("B")) * Int(5) / Int(100))
                        / Int(2),
                        Gtxn[5].receiver() == Addr(self.fee_addr2),
                        Gtxn[5].amount()
                        == (App.localGet(Int(1), Bytes("B")) * Int(5) / Int(100))
                        / Int(2),
                    )
                ),
                # Reset the bidder's offer
                App.localPut(Int(1), Bytes("B"), Int(0)),
                self.bids_amount.put(self.bids_amount.get() - Int(1)),
                # Transfer NFT to a new owner
                Assert(
                    Gtxn[3].asset_receiver() == Txn.accounts[1],
                ),
                If(
                    self.owner_address.get() == Global.zero_address(),
                    Seq(
                        [
                            Assert(
                                Gtxn[3].sender() == Txn.sender(),
                            ),
                            Return(Int(1)),
                        ]
                    ),
                ),
                If(
                    self.owner_address.get() == Txn.sender(),
                    Seq(
                        [
                            Assert(Gtxn[3].sender() == self.escrow_address.get()),
                            # Reset ask offer
                            self.ask_price.put(Int(0)),
                            self.owner_address.put(Global.zero_address()),
                            Return(Int(1)),
                        ]
                    ),
                ),
                Return(Int(0)),
            ]
        )

    def on_create(self):
        return Seq(
            [
                self.nft_id.put(Btoi(Txn.application_args[0])),
                self.owner_address.put(Global.zero_address()),
                self.creator_address.put(Txn.accounts[1]),
                self.escrow_address.put(Global.zero_address()),
                self.ask_price.put(Int(0)),
                self.bids_amount.put(Int(0)),
                self.contract_version.put(Int(self.contract_version_value)),
                Return(Int(1)),
            ]
        )

    def on_closeout(self):
        return Seq(
            [
                Assert(self.bid_price.get() == Int(0)),
                Return(Int(1)),
            ]
        )

    def on_setup_escrow(self):
        return Seq(
            [
                Assert(
                    self.escrow_address.get() == Global.zero_address(),
                ),
                Return(Int(1)),
            ]
        )

    def on_configure(self):
        return Seq(
            [
                # Update escrow address after creating it
                Assert(
                    And(
                        Txn.sender() == self.creator_address.get(),
                        self.escrow_address.get() == Global.zero_address(),
                    )
                ),
                self.escrow_address.put(Txn.accounts[1]),
                Return(Int(1)),
            ]
        )

    def get_contract(self):
        return Seq(
            [
                If(Txn.application_id() == Int(0), self.on_create()),
                Assert(
                    And(
                        Txn.close_remainder_to() == Global.zero_address(),
                        Txn.rekey_to() == Global.zero_address(),
                    )
                ),
                Cond(
                    [Txn.on_completion() == OnComplete.OptIn, self.on_register()],
                    [
                        Txn.on_completion()
                        == Or(
                            OnComplete.UpdateApplication, OnComplete.DeleteApplication
                        ),
                        Return(Int(0)),
                    ],
                    [Txn.on_completion() == OnComplete.CloseOut, self.on_closeout()],
                    [Txn.application_args[0] == Bytes("B"), self.on_bid()],
                    [Txn.application_args[0] == Bytes("S"), self.on_set_price()],
                    [Txn.application_args[0] == Bytes("BN"), self.on_buy_now()],
                    [Txn.application_args[0] == Bytes("SN"), self.on_sell_now()],
                    [Txn.application_args[0] == Bytes("C"), self.on_configure()],
                ),
            ]
        )


def manager(fee_addr1: str, fee_addr2: str, contract_version_value: str):
    return ManagerContract(fee_addr1, fee_addr2, contract_version_value).get_contract()


if __name__ == "__main__":
    params = {
        "fee_addr1": "MQGCPHNUM2LOACZ35R3HGGKEKAPAWOSEGCHKQH5OPOH6A5U2WQD5SYKRTA",
        "fee_addr2": "KFGORPVVRSWB6264OAKTROWZJ2QKJLAGTH7CP2Y2447P5ZFNIVFEKY3TPY",
        "contract_version_value": "1",
    }

    # Overwrite params if sys.argv[1] is passed
    if len(sys.argv) > 1:
        params = parse_params(sys.argv[1], params)

    print(
        compileTeal(
            manager(
                params["fee_addr1"],
                params["fee_addr2"],
                params["contract_version_value"],
            ),
            Mode.Application,
            version=TEAL_VERSION,
        )
    )



"""
MIT License

Copyright (c) 2022 AlgoWorld

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import dataclasses
import sys

from pyteal import Addr, And, Cond, Global, Gtxn, Int, Mode, TxnType, compileTeal

from algoworld_contracts.common.utils import parse_params

"""
ASA to ASA Atomic Swapper
1. Offered ASA Opt-In
2. Offered ASA / Required ASA Swap
3. Close Swap
"""

TEAL_VERSION = 6

MAX_FEE = Int(1000)
OPTIN_FUNDING_AMOUNT = 210000

ASA_OPTIN_GSIZE = Int(2)
OPTIN_FEE = 0
ASA_OPTIN = 1

ASA_SWAP_GSIZE = Int(3)
OFFERED_ASA_XFER = 0
REQUESTED_ASA_XFER = 1
INCENTIVE_FEE = 2

CLOSE_SWAP_GSIZE = Int(3)
ASA_CLOSE = 0
SWAP_CLOSE = 1
PROOF = 2


@dataclasses.dataclass
class AsaToAsaSwapConfig:
    swap_creator: str
    offered_asa_id: int
    offered_asa_amount: int
    requested_asa_id: int
    requested_asa_amount: int
    incentive_fee_address: str
    incentive_fee_amount: int


def swapper(cfg: AsaToAsaSwapConfig):

    is_asa_optin = And(
        Global.group_size() == ASA_OPTIN_GSIZE,
        Gtxn[OPTIN_FEE].type_enum() == TxnType.Payment,
        Gtxn[ASA_OPTIN].type_enum() == TxnType.AssetTransfer,
    )

    is_asa_swap = And(
        Global.group_size() == ASA_SWAP_GSIZE,
        Gtxn[OFFERED_ASA_XFER].type_enum() == TxnType.AssetTransfer,
        Gtxn[REQUESTED_ASA_XFER].type_enum() == TxnType.AssetTransfer,
        Gtxn[INCENTIVE_FEE].type_enum() == TxnType.Payment,
    )

    is_close_swap = And(
        Global.group_size() == CLOSE_SWAP_GSIZE,
        Gtxn[ASA_CLOSE].type_enum() == TxnType.AssetTransfer,
        Gtxn[SWAP_CLOSE].type_enum() == TxnType.Payment,
        Gtxn[PROOF].type_enum() == TxnType.Payment,
    )

    return Cond(
        [is_asa_optin, asa_optin(cfg)],
        [is_asa_swap, asa_swap(cfg)],
        [is_close_swap, close_swap(cfg)],
    )


def asa_optin(cfg: AsaToAsaSwapConfig):

    optin_fee_precondition = And(
        Gtxn[OPTIN_FEE].fee() <= MAX_FEE,
        Gtxn[OPTIN_FEE].rekey_to() == Global.zero_address(),
        Gtxn[OPTIN_FEE].close_remainder_to() == Global.zero_address(),
    )

    asa_optin_precondition = And(
        Gtxn[ASA_OPTIN].fee() <= MAX_FEE,
        Gtxn[ASA_OPTIN].rekey_to() == Global.zero_address(),
        Gtxn[ASA_OPTIN].asset_sender() == Global.zero_address(),
        Gtxn[ASA_OPTIN].asset_close_to() == Global.zero_address(),
    )

    return And(
        optin_fee_precondition,
        asa_optin_precondition,
        Gtxn[OPTIN_FEE].sender() == Addr(cfg.swap_creator),
        Gtxn[OPTIN_FEE].receiver() == Gtxn[ASA_OPTIN].sender(),
        Gtxn[OPTIN_FEE].amount() >= Int(OPTIN_FUNDING_AMOUNT),
        Gtxn[ASA_OPTIN].xfer_asset() == Int(cfg.offered_asa_id),
        Gtxn[ASA_OPTIN].sender() == Gtxn[ASA_OPTIN].asset_receiver(),
        Gtxn[ASA_OPTIN].asset_amount() == Int(0),
    )


def asa_swap(cfg: AsaToAsaSwapConfig):

    offered_asa_xfer_precondition = And(
        Gtxn[OFFERED_ASA_XFER].fee() <= MAX_FEE,
        Gtxn[OFFERED_ASA_XFER].rekey_to() == Global.zero_address(),
        Gtxn[OFFERED_ASA_XFER].asset_sender() == Global.zero_address(),
        Gtxn[OFFERED_ASA_XFER].asset_close_to() == Global.zero_address(),
    )

    requested_asa_xfer_precondition = And(
        Gtxn[REQUESTED_ASA_XFER].asset_sender() == Global.zero_address()
    )

    return And(
        offered_asa_xfer_precondition,
        requested_asa_xfer_precondition,
        Gtxn[OFFERED_ASA_XFER].xfer_asset() == Int(cfg.offered_asa_id),
        Gtxn[OFFERED_ASA_XFER].asset_amount() == Int(cfg.offered_asa_amount),
        Gtxn[REQUESTED_ASA_XFER].xfer_asset() == Int(cfg.requested_asa_id),
        Gtxn[REQUESTED_ASA_XFER].asset_amount() == Int(cfg.requested_asa_amount),
        Gtxn[OFFERED_ASA_XFER].asset_receiver() == Gtxn[REQUESTED_ASA_XFER].sender(),
        Gtxn[REQUESTED_ASA_XFER].asset_receiver() == Addr(cfg.swap_creator),
        Gtxn[INCENTIVE_FEE].receiver() == Addr(cfg.incentive_fee_address),
        Gtxn[INCENTIVE_FEE].sender() == Gtxn[REQUESTED_ASA_XFER].sender(),
        Gtxn[INCENTIVE_FEE].amount() == Int(cfg.incentive_fee_amount),
    )


def close_swap(cfg: AsaToAsaSwapConfig):

    asa_close_precondition = And(
        Gtxn[ASA_CLOSE].fee() <= MAX_FEE,
        Gtxn[ASA_CLOSE].rekey_to() == Global.zero_address(),
        Gtxn[ASA_CLOSE].asset_sender() == Global.zero_address(),
    )

    swap_close_precondition = And(
        Gtxn[SWAP_CLOSE].fee() <= MAX_FEE,
        Gtxn[SWAP_CLOSE].rekey_to() == Global.zero_address(),
    )

    return And(
        asa_close_precondition,
        swap_close_precondition,
        Gtxn[ASA_CLOSE].xfer_asset() == Int(cfg.offered_asa_id),
        Gtxn[ASA_CLOSE].asset_receiver() == Addr(cfg.swap_creator),
        Gtxn[ASA_CLOSE].asset_close_to() == Addr(cfg.swap_creator),
        Gtxn[SWAP_CLOSE].receiver() == Addr(cfg.swap_creator),
        Gtxn[SWAP_CLOSE].close_remainder_to() == Addr(cfg.swap_creator),
        Gtxn[PROOF].sender() == Addr(cfg.swap_creator),
        Gtxn[PROOF].receiver() == Addr(cfg.swap_creator),
        Gtxn[PROOF].amount() == Int(0),
    )


def compile_stateless(program):
    return compileTeal(program, Mode.Signature, version=TEAL_VERSION)


if __name__ == "__main__":
    params = {
        "swap_creator": "2ILRL5YU3FZ4JDQZQVXEZUYKEWF7IEIGRRCPCMI36VKSGDMAS6FHSBXZDQ",
        "offered_asa_id": 42,
        "offered_asa_amount": 1,
        "requested_asa_id": 69,
        "requested_asa_amount": 1,
        "incentive_fee_address": "RJVRGSPGSPOG7W3V7IMZZ2BAYCABW3YC5MWGKEOPAEEI5ZK5J2GSF6Y26A",
        "incentive_fee_amount": 10_000,
    }

    # Overwrite params if sys.argv[1] is passed
    if len(sys.argv) > 1:
        params = parse_params(sys.argv[1], params)

    print(compile_stateless(swapper(AsaToAsaSwapConfig(**params))))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['algoworld_contracts',
 'algoworld_contracts.auction',
 'algoworld_contracts.auction.helpers',
 'algoworld_contracts.common',
 'algoworld_contracts.swapper']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML==6.0', 'py-algorand-sdk==1.13.1', 'pyteal==0.10.0']

setup_kwargs = {
    'name': 'algoworld-contracts',
    'version': '1.2.4',
    'description': 'AlgoWorld Smart Contracts and Signatures',
    'long_description': '<p align="center"><a  href="https://twitter.com/algoworld_nft/status/1450608110268211203"><img  width=100%  src="https://i.imgur.com/cGWGmxa.png"  alt="687474703a2f2f6936332e74696e797069632e636f6d2f333031336c67342e706e67"  border="0" /></a></p>\n\n<p align="center">\n    <a href="https://algorand.com"><img src="https://img.shields.io/badge/Powered by-Algorand-blue.svg" /></a>\n    <a href="https://algoworld.io"><img src="https://img.shields.io/badge/AlgoWorld-Website-pink.svg" /></a>\n    <a href="https://algoworldexplorer.io"><img src="https://img.shields.io/badge/AlgoWorldExplorer-Platform-red.svg" /></a>\n    <a><img src="https://visitor-badge.glitch.me/badge?page_id=AlgoWorldNFT.algoworld-contracts&right_color=green" /></a>\n    <a href="https://github.com/AlgoWorldNFT/algoworld-contracts/actions/workflows/ci.yaml"><img src="https://github.com/AlgoWorldNFT/algoworld-contracts/actions/workflows/ci.yaml/badge.svg" /></a>\n    <a href="https://codecov.io/gh/AlgoWorldNFT/algoworld-contracts"><img src="https://codecov.io/gh/AlgoWorldNFT/algoworld-contracts/branch/main/graph/badge.svg?token=2O1VAOJCUD"  /></a>\n</p>\n\n## 📃 About\n\nThe following repository hosts the source codes for:\n- `AlgoWorld Swapper`\'s algorand smart signatures.\n- `AlgoWorldExplorer`\'s stateful smart contracts for card auctions and card trading. Modified version of [OpenNFT](https://github.com/ulamlabs/OpenNFT)\'s smart contracts.\n\n_**⚠️ NOTE: These contracts are not formally audited by accredited third parties. However, contracts are a basis for certain functionality on the AlgoWorldExplorer.io platform and were created in collaboration with Solution Architect from Algorand (credits @cusma). Code is provided under MIT license.**_\n\n## Prerequisites\n\n-   [poetry](https://python-poetry.org/)\n-   [pre-commit](https://pre-commit.com/)\n-   [Algorand Sandbox](https://github.com/algorand/sandbox)\n-   [Docker](https://www.docker.com/)\n\n## 🚀 Overview\n\nAlgoWorld currently offers stateful contracts used for auction trading on AlgoWorldExplorer and several smart signatures used for swapping on AlgoWorld Swapper.\n\n### Swapper\n\nThere are two different types of smart signatures available:\n\n- [ASA to ASA swap | 🎴↔️🎴](src/swapper/asa_to_asa_swapper.py): <br> Allows performing a swap of any single ASA of specified amount to any other single ASA of specified amount.\n\n- [ASAs to ALGO swap | 🎴🎴🎴↔️💰](src/swapper/asas_to_algo_swapper.py): <br> Allows performing a swap of multiple ASAs of specified amount to ALGO of specified amount.\n\n### Auction\n\nA set of stateful smart contracts for card auctions and card trading:\n- [ASAs for ALGO with trades and bidding | 🎴💰🔨](src/auction/manager.py): <br> Allows trading ASA via auctions, selling or purchasing directly.\n\n## ⚙️ Installation\n\nThis section assumes that poetry and pre-commit are installed and executed from the root folder of this repository.\n\n1. Clone the repo\n\n```bash\ngit clone https://github.com/AlgoWorldNFT/algoworld-contracts\n```\n\n2. Install python requirements\n\n```bash\npoetry install # install all dependencies\npoetry shell # activate virtual env\n```\n\n(OPTIONAL) 3. Configure `pre-commit` hooks\n\n```bash\npre-commit install\n```\n\nIf you are not going to setup `pre-commit` locally, there is a Github Actions plugin that will autoformat your branch if you are opening a PR with commits that contain un-formatted code.\n\n## 🧪 Testing\n\nTesting assumes that docker-compose is installed and available. Project is relying on `pytest-docker-compose` plugin that automatically boots up temporary algorand sandbox and destroys the containers after the tests are finished.\n\n```bash\n(.venv) pytest\n```\n\nYou can also include `[pytest]` into your commit message to trigger the test in CI pipeline on `push` action (on pr it is triggered automatically).\n\n## 🚧 Contribution guideline\n\nSee [`CONTRIBUTING.md`](CONTRIBUTING.md)\n\n## ⭐️ Stargazers\n\nSpecial thanks to everyone who forked or starred the repository ❤️\n\n[![Stargazers repo roster for @AlgoWorldNFT/algoworld-contracts](https://reporoster.com/stars/dark/AlgoWorldNFT/algoworld-contracts)](https://github.com/AlgoWorldNFT/algoworld-contracts/stargazers)\n\n[![Forkers repo roster for @AlgoWorldNFT/algoworld-contracts](https://reporoster.com/forks/dark/AlgoWorldNFT/algoworld-contracts)](https://github.com/AlgoWorldNFT/algoworld-contracts/network/members)\n',
    'author': 'AlgoWorld',
    'author_email': 'info@algoworld.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AlgoWorldNFT/algoworld-contracts',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
